#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <detect.h>
#include <math.h>

/*
 * Authors:
 *    Koen Van der Goten & Andre Vandierendonck
 *    Department of Experimental Psychology
 *    University of Ghent
 *    Henri Dunantlaan 2
 *    B-9000 Gent
 *    Phone: +32-(0)9-264 64 06 or +32-(0)9-264 64 37
 *    Fax: +32-(0)9-264 64 96
 *    E-mail: Koen.VanderGoten@rug.ac.be
 *            Andre.Vandierendonck@rug.ac.be
 * $Author: andre $
 * $Date: 96/07/24 16:49:16 $
 * $Revision: 1.0 $
 * $Log:	detect.c,v $
 * Revision 1.0  96/07/24  16:49:16  andre
 * Initial revision
 * 
 *
 * This program calculates indices of sensitivity and bias based
 * on several theories. All the functions used are part of a library
 * module, which is in a separate file.
 */

void intro(void);
void printhead(double ns, double n, double h1, double f1, double ns2,
    double n2, double h2, double f2, int selection, int extreme);
void gaussian(double ns, double n, double h1, double f1, int extreme);
void logistic(double ns, double n, double h1, double f1, int extreme);
void rectangular(double ns, double n, double h1, double f1, int extreme);
void other(double ns, double n, double h1, double f1, int extreme);
void roc(double ns, double n, double h1, double f1, double ns2, double n2,
    double h2, double f2, int extreme);
 
void main(int argc, char *argv[])
{
    double ns, n, ns2, n2, h1, f1, h2, f2;
    int extreme;
    int selection;

    if (argc==1)
	intro();
    if (argc>1)
    {
	if (strcasecmp("a", argv[1])==0)   selection=0;
	else if (strcasecmp("g", argv[1])==0)   selection=1;
	else if (strcasecmp("l", argv[1])==0)   selection=2;
	else if (strcasecmp("r", argv[1])==0)   selection=3;
	else if (strcasecmp("o", argv[1])==0)   selection=4;
	else if (strcasecmp("roc", argv[1])==0) selection=5;
	else
	{
	    fprintf(stderr, "\nunknown selection in argument 1:");
	    fprintf(stderr, " %s\n\n", argv[1]);
	    exit(0);
	}
    }
    if (selection>=0 && selection<=4 && argc!=7)
    {
	detect_ver();
	fprintf(stderr, "\nThe selection in the first argument ");
	fprintf(stderr, "requires a total of 6 arguments:\n");
	fprintf(stderr, "   1. selection\n");
	fprintf(stderr, "   2. flag for extreme proportions\n");
	fprintf(stderr, "   3. number of signal presentations\n");
	fprintf(stderr, "   4. number of noise presentations\n");
	fprintf(stderr, "   5. number of hits\n");
	fprintf(stderr, "   6. number of false alarms\n");
	exit(0);
    }

    if (selection==5 && argc!=11)
    {
	detect_ver();
	fprintf(stderr, "\nThe selection in the first argument ");
	fprintf(stderr, "requires a total of 8 arguments:\n");
	fprintf(stderr, "   1. selection\n");
	fprintf(stderr, "   2. flag for extreme proportions\n");
	fprintf(stderr, "   3. point 1: number of signal presentations\n");
	fprintf(stderr, "   4. point 1: number of noise presentations\n");
	fprintf(stderr, "   5. point 1: number of hits\n");
	fprintf(stderr, "   6. point 1: number of false alarms\n");
	fprintf(stderr, "   7. point 2: number of signal presentations\n");
	fprintf(stderr, "   8. point 2: number of noise presentations\n");
	fprintf(stderr, "   9. point 2: number of hits\n");
	fprintf(stderr, "  10. point 2: number of false alarms\n");
	exit(0);
    }

    if(strcasecmp("2n",argv[2])!=0 && strcasecmp("ll",argv[2])!=0 && argc>1)
    {
	detect_ver();
	fprintf(stderr,"\nArgument 2 should set flag");
	fprintf(stderr," for extreme proportions:\n");
	fprintf(stderr," - ll for log-linear rule\n");
	fprintf(stderr," - 2n for correction with");
	fprintf(stderr," 1\\2n\n");
	exit(0);
    }
    extreme = strcasecmp("2n", argv[2]);

    ns    = atof(argv[3]);
    n     = atof(argv[4]);
    h1    = atof(argv[5]);
    f1    = atof(argv[6]);

    if (selection==5)
    {
	ns2 = atof(argv[7]);
	n2 = atof(argv[8]);
	h2 = atof(argv[9]);
	f2 = atof(argv[10]);
    }

    printhead(ns, n, h1, f1, ns2, n2, h2, f2, selection, extreme);

    switch (selection) {
    case 0:
	gaussian(ns, n, h1, f1, extreme);
	logistic(ns, n, h1, f1, extreme);
	rectangular(ns, n, h1, f1, extreme);
	other(ns, n, h1, f1, extreme);
	break;
    case 1:
	gaussian(ns, n, h1, f1, extreme);
	break;
    case 2:
	logistic(ns, n, h1, f1, extreme);
	break;
    case 3:
	rectangular(ns, n, h1, f1, extreme);
	break;
    case 4:
	other(ns, n, h1, f1, extreme);
	break;
    case 5:
	roc(ns, n, h1, f1, ns2, n2, h2, f2, extreme);
    }
}

/* intro -- print help information */
void intro(void)
{
    detect_ver();
    fprintf(stderr,
	"- arg 1 determines which calculations will be performed:\n");
    fprintf(stderr,"\tG:   gives calculations under GAUSSIAN assumptions\n");
    fprintf(stderr,"\tL:   gives calculations under LOGISTIC assumptions\n");
    fprintf(stderr,"\tR:   gives calculations under RECTANGULAR assumptions\n");
    fprintf(stderr,"\tO:   gives OTHER than previous calculations\n");
    fprintf(stderr,"\tA:   gives ALL preceding calculations\n");
    fprintf(stderr,"\tROC: gives calculations based on ROC (2 points)\n");
    fprintf(stderr,
	"- arg 2 sets flag for calculation of extreme proportions:\n");
    fprintf(stderr,"\t2N:  correction with 2N\n");
    fprintf(stderr,"\tLL:  log-linear rule\n");
    fprintf(stderr,"- arg 3 number of signal presentations\n");
    fprintf(stderr,"- arg 4 number of noise presentations\n");
    fprintf(stderr,"- arg 5 number of hits\n");
    fprintf(stderr,"- arg 6 number of false alarms\n");
    fprintf(stderr,"- in case of ROC in arg 1:\n");
    fprintf(stderr,
	"\targ 7 number of signal presentations for the 2nd point\n");
    fprintf(stderr,"\targ 8 number of noise presentations for the 2nd point\n");
    fprintf(stderr,"\targ 9 number of hits for the 2nd point\n");
    fprintf(stderr,"\targ 10 number of false alarms for the 2nd point\n");
    fprintf(stderr,"Check the article for more information!\n");
    exit(0);
}

/* printhead -- print header of output */
void printhead(double ns, double n, double h1, double f1, double ns2,
double n2, double h2, double f2, int selection, int extreme)
{
    int junk=0;

    if (selection<5)
    {
	printf("\nHit ratio         : %3.lf / %3.lf\n", h1, ns);
	printf("False alarm ratio : %3.lf / %3.lf\n", f1, n);
    }
    else if (selection==5)
    {
	printf("\nHit ratio         : %3.lf / %3.lf, %3.lf / %3.lf\n",
	    h1, ns, h2, ns2);
	printf("False alarm ratio : %3.lf / %3.lf, %3.lf / %3.lf\n",
	    f1, n, f2, n2);
    }

    if (!extreme)
    {
	if  (h1==0. || f1==0. || h1==ns || f1==n)
	    junk++;
	if (selection==5)
	{
	    if (h2==0. || f2==0. || h2==ns2 || f2==n2)
		junk++;
	}
	if (junk)
	    printf
		("2N-Adjustment has been done for ratio of 0.0 or of 1.0\n\n");
    }
    if (extreme)
	printf("Adjustment with log-linear rule has been done \n\n");
}

/* gaussian -- calculate and print all gaussian indices */
void gaussian(double ns, double n, double h1, double f1, int extreme)
{
    double dprime, c, cprime, beta_g, logbeta_g;

    dprime    = sens_g(ns, n, h1, f1, extreme);
    c         = loc_g(ns, n, h1, f1, extreme);
    cprime    = relloc_g(ns, n, h1, f1, extreme);
    beta_g    = like_g(ns, n, h1, f1, extreme);
    logbeta_g = loglike_g(ns, n, h1, f1, extreme);

    printf("\nGaussian Distribution\n");
    printf("d'           = %11.6lf\n", dprime);
    printf("c            = %11.6lf\n", c);
    if (cprime==HUGE)
    {
	printf("c'           =    undefined\n");
	printf("Beta(G)      =    undefined\n");
	printf("log(Beta(G)) =    undefined\n");
    }
    else
    {
	printf("c'           = %11.6lf\n", cprime);
	printf("Beta(G)      = %11.6lf\n", beta_g);
	printf("log(Beta(G)) = %11.6lf\n", logbeta_g);
    }
}

/* logistic -- calculate and print all logistic indices */
void logistic(double ns, double n, double h1, double f1, int extreme)
{
    double logalfa, logb, alfa, b, beta_l, logbeta_l, bprime, b_hodos, bgrier;

    logalfa      = sens_l(ns, n, h1, f1, extreme);
    logb         = logloc_l(ns, n, h1, f1, extreme);
    alfa         = sens_alfa(ns, n, h1, f1, extreme);
    b            = loc_l(ns, n, h1, f1, extreme);
    beta_l       = like_l(ns, n, h1, f1, extreme);
    logbeta_l    = loglike_l(ns, n, h1, f1, extreme);
    bprime       = relloc_l(ns, n, h1, f1, extreme);
    b_hodos      = hodos(ns, n, h1, f1, extreme);
    bgrier       = grier(ns, n, h1, f1, extreme);

    printf("\nLogistic Distribution\n");
    printf("logalfa      = %11.6lf\n", logalfa);
    printf("log(b)       = %11.6lf\n", logb);
    printf("alfa         = %11.6lf\n", alfa);
    printf("b            = %11.6lf\n", b);
    if (bprime==HUGE)
    {
	printf("b'           =    undefined\n");
	printf("Beta(L)      =    undefined\n");
	printf("log(Beta(L)) =    undefined\n");
	printf("B'(H)        =    undefined\n");
	printf("B''          =    undefined\n");
    }
    else
    {
	printf("b'           = %11.6lf\n", bprime);
	printf("Beta(L)      = %11.6lf\n", beta_l);
	printf("log(Beta(L)) = %11.6lf\n", logbeta_l);
	printf("B'(H)        = %11.6lf\n", b_hodos);
	printf("B''          = %11.6lf\n", bgrier);
    }
}

/* rectangular -- calculate and print rectangular based indices */
void rectangular(double ns, double n, double h1, double f1, int extreme)
{
    double q, p_c, k, kprime;

    q            = sens_q(ns, n, h1, f1, extreme);
    p_c          = sens_pc(ns, n, h1, f1, extreme);
    k            = loc_k(ns, n, h1, f1, extreme);
    kprime       = relloc_k(ns, n, h1, f1, extreme);

    printf("\nRectangular Distributions\n");
    printf("q            = %11.6lf\n", q);
    printf("p(c)         = %11.6lf\n", p_c);
    printf("k            = %11.6lf\n", k);
    printf("k'           = %11.6lf\n", kprime);
}

/* other -- calculate and print other indices */
void other(double ns, double n, double h1, double f1, int extreme)
{
    double lor, yule, aprime;

    lor          = sens_lor(ns, n, h1, f1, extreme);
    yule         = sens_yule(ns, n, h1, f1, extreme);
    aprime       = sens_np(ns, n, h1, f1, extreme);

    printf("\n`Nonparametric'\n");
    printf("A'           = %11.6lf\n", aprime);
    printf("\nSensitivity measures proposed by Swets (1986)\n");
    printf("LOR          = %11.6lf\n", lor);
    printf("Yule's Q     = %11.6lf\n", yule);
}

/* roc -- calculate and print ROC-indices */
void roc(double ns, double n, double h1, double f1, double ns2, double n2,
double h2, double f2, int extreme)
{
    double s, deltam, da, de, az;

    s =       roc_slope(ns, n, h1, f1, ns2, n2, h2, f2, extreme);
    deltam =  delta_m(ns, n, h1, f1, ns2, n2, h2, f2, extreme);
    da =      roc_da(ns, n, h1, f1, ns2, n2, h2, f2, extreme);
    de =      roc_de(ns, n, h1, f1, ns2, n2, h2, f2, extreme);
    az =      roc_az(ns, n, h1, f1, ns2, n2, h2, f2, extreme);

    printf("\nROC measures\n");
    printf("slope        = %11.6lf\n", s);
    printf("delta m      = %11.6lf\n", deltam);
    printf("da           = %11.6lf\n", da);
    printf("d'e          = %11.6lf\n", de);
    printf("Az           = %11.6lf\n", az);
}
