#include <detect.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

/*
 * Authors:
 *    Koen Van der Goten & Andre Vandierendonck
 *    Department of Experimental Psychology
 *    University of Ghent
 *    Henri Dunantlaan 2
 *    B-9000 Gent
 *    Phone: +32-(0)9-264 64 06 or +32-(0)9-264 64 37
 *    Fax: +32-(0)9-264 64 96
 *    E-mail: Koen.VanderGoten@rug.ac.be
 *            Andre.Vandierendonck@rug.ac.be
 * $Author: andre $
 * $Date: 96/07/24 16:51:36 $
 * $Revision: 1.0 $
 * $Log:	detect.c,v $
 * Revision 1.0  96/07/24  16:51:36  andre
 * Initial revision
 * 
 *
 * ==================================================================== *
 * This is the c-library module described in Van der Goten, K. \&	*
 * Vandierendonck, A. Software to calculate measures of sensitivity	*
 * and response bias based on detection theory and threshold theory.	*
 * References for each of the measures and indices includes are given	*
 * in this article.							*
 *									*
 * Characteristics of the programming approach:				*
 * - standard calling sequences to all functions			*
 *   + for the simple sensitivity and bias measures:			*
 *	number of noise+signal trials (double ns)			*
 *	number of noise trials (double n)				*
 *	number of hits (double h1)					*
 *	number of false alarms (double f1)				*
 *	flag for handling extreme data (int extreme; 0/FALSE, 1/2n-rule	*
 *		(TRUE, log-linear rule, Hautus)				*
 *   + for the roc-based measures:					*
 *	number of noise+signal trials (double ns)			*
 *	number of noise trials (double n)				*
 *	number of hits in first categorization (double h1)		*
 *	number of false alarms in first categorization (double f1)	*
 *	number of noise+signal trials (double ns2)			*
 *	number of noise trials (double n2)				*
 *	number of hits in second categorization (double h2)		*
 *	number of false alarms in second categorization (double f2)	*
 *	flag for handling extreme data (int extreme; 0/FALSE, 1/2n-rule	*
 *		(TRUE, log-linear rule, Hautus)				*
 * - names of the functions follow a transparant scheme			*
 * - modular approach							*
 *   e.g., function ratio calculates the proportion of hits or false	*
 *   alarms and applies the requested adaptation for extreme values,	*
 *   and is called by every other function in the package. It is also	*
 *   the only function that may cause an error message and exit		*
 *									*
 * One subroutine was added for purposes of control: detect_ver prints	*
 * the revision number and the date to the standard error output.	*
 *									*
 * CONTENTS of the PACKAGE:						*
 * -------------------------------------------------------------------- *
 * Distribution			Measure			Function	*
 * -------------------------------------------------------------------- *
 * Gaussian			d'			sens_g		*
 *				c			loc_g		*
 *				d'			relloc_g	*
 *				beta_G			like_g		*
 *				log(beta_G)		loglike_g	*
 * -------------------------------------------------------------------- *
 * Logistic			alpha			sens_alpha	*
 *				b			loc_l		*
 *				log(alpha)		sens_l		*
 *				log(b)			logloc_l	*
 *				b'			relloc_l	*
 *				beta_L			like_l		*
 *				log(beta_L)		loglike_l	*
 *				B'_H			hodos		*
 *				B"			grier		*
 * -------------------------------------------------------------------- *
 * Rectangular			q			sens_q		*
 *				P(c)			sens_pc		*
 *				k			loc_k		*
 *				k'			relloc_k	*
 * -------------------------------------------------------------------- *
 * "Nonparametric"		A'			sens_np		*
 * -------------------------------------------------------------------- *
 * Other ones			LOR			sens_lor	*
 *				Yule's Q		sens_yule	*
 * -------------------------------------------------------------------- *
 * ROC				slope			roc_slope	*
 *				Delta_m			delta_m		*
 *				d_a			roc_da		*
 *				d'_e			roc_de		*
 *				A_z			roc_az		*
 * -------------------------------------------------------------------- *
 * Print revision info					detect_ver	*
 * -------------------------------------------------------------------- *
 *									*
 * Copyright (C) Koen Van der Goten and Andre Vandierendonck.		*
 * Everyone is permitted to copy and distribute verbatim copies of this	*
 * document, but changing is not allowed. The software is provided	*
 * as-is without any warranty of any kind. Selling or sublicensing of	*
 * this product is not allowed.						*
 * ==================================================================== *
 */

/* detect_ver -- print the revision number and the last modification
 * date of the package */
void detect_ver(void)
{
    fprintf(stderr,
	"Module detect: $Revision: 1.0 $, $Date: 96/07/24 16:51:36 $\n");
}

/* ratio -- calculate the proportion x/y corrected by log-linear rule
 * if extreme is TRUE, correct by 1/2n rule otherwise */
double ratio(double x, double y, int extreme)
{
    double r;

    if (x < 0. || y <= 0. || x > y)	/* invalid parameter */
    {
    	fprintf(stderr, "Invalid parameter, exiting\n");
    	exit(1);
    }

    if (extreme)		/* log-linear rule */
    {
    	x += .5;	/* each entry of 2x2 contingency table +.5 */
    	y += 1.;	/* row and column totals +1. */
    }

    r = x/y;
    if (r == 0.)			/* 1/2n -rule */
    	r = 1./(2.*y);
    else if (r == 1.)
    	r = 1. - 1./(2.*y);

    return(r);
}

/* odeh -- Approximation of the inverse normal distribution function
 * proposed by Odeh & Evans (1974) in Applied Statistics.
 * For discussion see Brophy (1985) in Behavior Research
 * Methods, Instruments & Computers. */
double odeh(double p)
{
    double r, y, z;
 
    r = p;
    if (p > .5)
	r = 1 - p;
    y = sqrt(-2.*log(r));
    z = y - ((((4.53642210148E-5 * y + .0204231210245) * y + .342242088547) *
        y + 1) * y + .322232431088) / ((((.0038560700634 * y + .10353775285) * 
        y + .531103462366) * y + .588581570495) * y + .099348462606);
    if (p < .5)
	z = -z;
    return(z);
}

/* sens_g -- Gaussian measure of sensitivity (d') */
double sens_g(double ns, double n, double h1, double f1, int extreme)
{
    double zh, zf, d;
 
    h1 = ratio(h1, ns, extreme);
    f1 = ratio(f1, n, extreme);
    zh = (odeh(h1));
    zf = (odeh(f1));
    d = zh - zf;
    return(d);
}

/* loc_g -- Gaussian estimate of bias: location of criterion (c) */
double loc_g(double ns, double n, double h1, double f1, int extreme)
{
    double zh, zf, c;

    h1 = ratio(h1, ns, extreme);
    f1 = ratio(f1, n, extreme);
    zh = (odeh(h1));
    zf = (odeh(f1));
    c = -.5 * (zf + zh);
    return(c);
}

/* relloc_g -- Gaussian estimate of bias: relative location of criterion (c') */
double relloc_g(double ns, double n, double h1, double f1, int extreme)
{
    double d, c, cprime;

    if (h1/ns == f1/n)	/* measure is undefined */
    	return(HUGE);

    d = sens_g(ns, n, h1, f1, extreme);
    c = loc_g(ns, n, h1, f1, extreme);
    cprime = c/d;
    return(cprime);
}

/* like_g -- Gaussian likelihood ratio (beta_G) */
double like_g(double ns, double n, double h1, double f1, int extreme)
{
    double b, zh, zf, hh, hf;
 
    if (h1/ns == f1/n)	/* measure is undefined */
    	return(HUGE);

    h1 = ratio(h1, ns, extreme);
    f1 = ratio(f1, n, extreme);
    zh = odeh(h1);
    zf = odeh(f1);

    hh = (1./(sqrt(2. * M_PI))) * exp(-.5 * zh * zh);
    hf = (1./(sqrt(2. * M_PI))) * exp(-.5 * zf * zf);

    b = (hh / hf) * (ns / n);
    if (f1>h1)
	b = 1./b;
    return(b);
}

/* loglike_g -- Logarithm of gaussian likelihood ratio log(beta_G) */
double loglike_g(double ns, double n, double h1, double f1, int extreme)
{
    double b, zh, zf, hh, hf;
 
    if (h1/ns == f1/n)	/* measure is undefined */
    	return(HUGE);

    b = like_g(ns, n, h1, f1, extreme);
    b = log(b);
    return(b);
}

/* sens_l -- Logarithm of sensitivity measure based on choice theory
 * log(alpha) */
double sens_l(double ns, double n, double h1, double f1, int extreme)
{
    double logalfa;
 
    h1 = ratio(h1, ns, extreme);
    f1 = ratio(f1, n, extreme);
    logalfa = (.5 * log(h1/(1. - h1)) -.5 * log(f1/(1. - f1)));
    return(logalfa);
}

/* sens_alfa -- Choice theory measure of sensitivity, alpha */
double sens_alfa(double ns, double n, double h1, double f1, int extreme)
{
    double alfa;
 
    h1 = ratio(h1,ns, extreme);
    f1 = ratio(f1,n, extreme);
    alfa = sqrt((h1 * (1. - f1))/((1. - h1) * f1));
    return(alfa);
}

/* loc_l -- Choice theory measure of bias, b */
double loc_l(double ns, double n, double h1, double f1, int extreme)
{
    double b;
 
    h1 = ratio(h1, ns, extreme);
    f1 = ratio(f1, n, extreme);
    b = sqrt((1. - h1) * (1. - f1)/(h1 * f1));
    return(b);
}


/* logloc_l -- Logarithm of choice theory measure of bias, log(b) */
double logloc_l(double ns, double n, double h1, double f1, int extreme)
{
    double logb;

    h1 = ratio(h1, ns, extreme);
    f1 = ratio(f1, n, extreme);
    logb = -.5 * (log(h1/(1. - h1)) + log(f1/(1. - f1)));
    return(logb);
}

/* relloc_l -- Choice theory measure of relative location, b' */
double relloc_l(double ns, double n, double h1, double f1, int extreme)
{
    double logalfa, logb, bprime;
 
    if (h1/ns == f1/n)	/* measure is undefined */
    	return(HUGE);

    logalfa = sens_l(ns, n, h1, f1, extreme);
    logb = logloc_l(ns, n, h1, f1, extreme);
    bprime = logb/(2. * logalfa);
    if (logalfa < 0.)
	bprime = -bprime;
    return(bprime);
}

/* like_l -- Logistic based likelihood ratio, beta_L */
double like_l(double ns, double n, double h1, double f1, int extreme)
{
    double beta_l;
 
    if (h1/ns == f1/n)	/* measure is undefined */
    	return(HUGE);

    h1 = ratio(h1, ns, extreme);
    f1 = ratio(f1, n, extreme);
    if ( h1>f1 )
	beta_l = h1 * (1. - h1)/(f1 * (1. - f1));
    else
	beta_l = f1 * (1. - f1)/(h1 * (1. - h1));
    return(beta_l);
}

/* loglike_l -- Logarithm of logistic likelihood ratio, log(beta_L) */
double loglike_l(double ns, double n, double h1, double f1, int extreme)
{
    double beta_l;
 
    if (h1/ns == f1/n)	/* measure is undefined */
    	return(HUGE);

    beta_l = log(like_l(ns, n, h1, f1, extreme));
    return(beta_l);
}

/* hodos -- formula Hodos as extended by Singh and Karney */
double hodos(double ns, double n, double h1, double f1, int extreme)
{
    double x, y;
 
    if (h1/ns == f1/n)	/* measure is undefined */
    	return(HUGE);

    h1 = ratio(h1, ns, extreme);
    f1 = ratio(f1, n, extreme);
    if (h1 < f1)
    {
	x = h1;
	y = f1;
    }
    else
    {
	x = f1;
	y = h1;
    }
    if (x + y <= 1.)
	return(1. - ((x * (1. - x))/(y * (1. - y))));
    else
	return((y * (1. - y))/(x * (1. - x)) - 1.);
}

/* grier -- Adaptation of Hodos' bias formula proposed by Grier */
double grier(double ns, double n, double h1, double f1, int extreme)
{
    double x, y, b;
 
    if (h1/ns == f1/n)	/* measure is undefined */
    	return(HUGE);

    h1 = ratio(h1, ns, extreme);
    f1 = ratio(f1, n, extreme);
    if (h1 < f1)
    {
	x = h1;
	y = f1;
    }
    else
    {
	x = f1;
	y = h1;
    }
    b = (y * (1. - y) - x * (1. - x)) / (y * (1. - y) + x * (1. - x));
    return(b);
}

/* sens_q -- Single high threshold theory sensitivity index q */
double sens_q(double ns, double n, double h1, double f1, int extreme)
{
    double q;
 
    h1 = ratio(h1, ns, extreme);
    f1 = ratio(f1, n, extreme);
    q = (h1 - f1)/(1. - f1);
    return(q);
}

/* sens_pc -- Double high threshold theory measure of sensitivity, P(c) */
double sens_pc(double ns, double n, double h1, double f1, int extreme)
{
    double p_c;
 
    h1 = ratio(h1, ns, extreme);
    f1 = ratio(f1, n, extreme);
    p_c = .5 * (1. - (f1 - h1));
    return(p_c);
}

/* loc_k -- Double high threshold theory measure of location, k */
double loc_k(double ns, double n, double h1, double f1, int extreme)
{
    double k;

    h1 = ratio(h1, ns, extreme);
    f1 = ratio(f1, n, extreme);
    k = .5 * (1. - (f1 + h1));
    return(k);
}

/* relloc_k -- Double high threshold theory measure of relative location, k' */
double relloc_k(double ns, double n, double h1, double f1, int extreme)
{
    double kprime;

    h1 = ratio(h1, ns, extreme);
    f1 = ratio(f1, n, extreme);
    kprime = 1./(1. + (f1/(1. - h1)));
    return(kprime);
}

/* sens_lor -- LOR measure */
double sens_lor(double ns, double n, double h1, double f1, int extreme)
{
    double lor;

    h1 = ratio(h1, ns, extreme);
    f1 = ratio(f1, n, extreme);
    lor = log( (h1 * (1. - f1)) / (f1 * (1. - h1)) );
    return(lor);
}

/* sens_yule -- Yule's Q measure of sensitivity */
double sens_yule(double ns, double n, double h1, double f1, int extreme)
{
    double yule;
 
    h1 = ratio(h1, ns, extreme);
    f1 = ratio(f1, n, extreme);
    yule = (h1 - f1)/(h1 - 2. * f1 * h1 + f1);
    return(yule);
}

/* sens_np -- A-prime "nonparametric" measure of sensitivity */
double sens_np(double ns, double n, double h1, double f1, int extreme)
{
    double x, y;
    double aprime;

    y = ratio(h1, ns, extreme);
    x = ratio(f1, n, extreme);
    if (y >= x)                     /* p(hits) >= p(fa) */
	aprime = .5 + (y - x + (x - y) * (x - y))/(4 * y * (1. - x));
    else
	aprime = .5 - (x - y + (x - y) * (x - y))/(4 * x * (1. - y));
    return(aprime);
}

/* roc_slope -- Calculate the slope of the ROC curve when two points
 * are given */
double roc_slope(double ns, double n, double h1, double f1, double ns2,
double n2, double h2, double f2, int extreme)
{
    double zh1, zf1, zh2, zf2, s;

    h1 = ratio(h1, ns, extreme);
    f1 = ratio(f1, n, extreme);
    h2 = ratio(h2, ns2, extreme);
    f2 = ratio(f2, n2, extreme);
    zh1 = (odeh(h1));
    zf1 = (odeh(f1));
    zh2 = (odeh(h2));
    zf2 = (odeh(f2));
    s = fabs((zh2-zh1) / (zf2-zf1));
    return(s);
}

/* delta_m */
double delta_m(double ns, double n, double h1, double f1, double ns2,
double n2, double h2, double f2, int extreme)
{
    double zh1, zf1, zh2, zf2, s, deltam;

    h1 = ratio(h1, ns, extreme);
    f1 = ratio(f1, n, extreme);
    h2 = ratio(h2, ns2, extreme);
    f2 = ratio(f2, n2, extreme);
    zh1 = (odeh(h1));
    zf1 = (odeh(f1));
    zh2 = (odeh(h2));
    zf2 = (odeh(f2));
    s = fabs((zh2-zh1) / (zf2-zf1));
    deltam= zh1/s - zf1;
    return(deltam);
}
 
/* roc_da -- Index d_a */
double roc_da(double ns, double n, double h1, double f1, double ns2,
double n2, double h2, double f2, int extreme)
{
    double zh1, zf1, zh2, zf2, s, da;

    h1 = ratio(h1, ns, extreme);
    f1 = ratio(f1, n, extreme);
    h2 = ratio(h2, ns2, extreme);
    f2 = ratio(f2, n2, extreme);
    zh1 = (odeh(h1));
    zf1 = (odeh(f1));
    zh2 = (odeh(h2));
    zf2 = (odeh(f2));
    s = fabs((zh2 - zh1) / (zf2 - zf1));
    da = sqrt(2./(1. + s * s)) * (zh1 - s * zf1);
    return(da);
}
 
/* roc_de -- Distance measure d'_e */
double roc_de(double ns, double n, double h1, double f1, double ns2,
double n2, double h2, double f2, int extreme)
{
    double zh1, zf1, zh2, zf2, s, de;

    h1 = ratio(h1, ns, extreme);
    f1 = ratio(f1, n, extreme);
    h2 = ratio(h2, ns2, extreme);
    f2 = ratio(f2, n2, extreme);
    zh1 = (odeh(h1));
    zf1 = (odeh(f1));
    zh2 = (odeh(h2));
    zf2 = (odeh(f2));
    s = fabs((zh2 - zh1) / (zf2 - zf1));
    de = 2./(1. + s) * (zh1 - s * zf1);
    return(de);
}
 
/* roc_az -- Index A_z */
double roc_az(double ns, double n, double h1, double f1, double ns2,
double n2, double h2, double f2, int extreme)
{
    double da, az;

    da = roc_da(ns, n, h1, f1, ns2, n2, h2, f2, extreme);
    az = (1./(sqrt(2. * M_PI))) * exp(-.5 * da * da);
    return(az);
}
