unit dtctunit;

{*
 * Authors:
 *    Koen Van der Goten & Andre Vandierendonck
 *    Department of Experimental Psychology
 *    University of Ghent
 *    Henri Dunantlaan 2
 *    B-9000 Gent
 *    Phone: +32-(0)9-264 64 06 or +32-(0)9-264 64 37
 *    Fax: +32-(0)9-264 64 96
 *    E-mail: Koen.VanderGoten@rug.ac.be
 *            Andre.Vandierendonck@rug.ac.be
 * $Author: andre $
 * $Date: 1996/07/24 16:28:47 $
 * $Revision: 1.0 $
 * $Log: dtctunit.pas,v $
 * Revision 1.0  1996/07/24 16:28:47  andre
 * Initial revision
 *
 *
 * Translated to Pascal by Hans Vandierendonck
 *
 * ==================================================================== *
 * This is the c-library module described in Van der Goten, K. \&       *
 * Vandierendonck, A. Software to calculate measures of sensitivity     *
 * and response bias based on detection theory and threshold theory.    *
 * References for each of the measures and indices includes are given   *
 * in this article.                                                     *
 *                                                                      *
 * Characteristics of the programming approach:                         *
 * - standard calling sequences to all functions                        *
 *   + for the simple sensitivity and bias measures:                    *
 *      number of noise+signal trials (double ns)                       *
 *      number of noise trials (double n)                               *
 *      number of hits (double h1)                                      *
 *      number of false alarms (double f1)                              *
 *      flag for handling extreme data (int extreme; 0/FALSE, 1/2n-rule *
 *              (TRUE, log-linear rule, Hautus)                         *
 *   + for the roc-based measures:                                      *
 *      number of noise+signal trials (double ns)                       *
 *      number of noise trials (double n)                               *
 *      number of hits in first categorization (double h1)              *
 *      number of false alarms in first categorization (double f1)      *
 *      number of noise+signal trials for second point (double ns2)     *
 *      number of noise trials for second point (double n2)		*
 *      number of hits in second categorization (double h2)             *
 *      number of false alarms in second categorization (double f2)     *
 *      flag for handling extreme data (int extreme; 0/FALSE, 1/2n-rule *
 *              (TRUE, log-linear rule, Hautus)                         *
 * - names of the functions follow a transparant scheme                 *
 * - modular approach                                                   *
 *   e.g., function ratio calculates the proportion of hits or false    *
 *   alarms and applies the requested adaptation for extreme values,    *
 *   and is called by every other function in the package. It is also   *
 *   the only function that may cause an error message and exit         *
 *                                                                      *
 * One subroutine was added for purposes of control: detect_ver prints  *
 * the revision number and the date to the standard error output.       *
 *                                                                      *
 * CONTENTS of the PACKAGE:                                             *
 * -------------------------------------------------------------------- *
 * Distribution                 Measure                 Function        *
 * -------------------------------------------------------------------- *
 * Gaussian                     d'                      sens_g          *
 *                              c                       loc_g           *
 *                              d'                      relloc_g        *
 *                              beta_G                  like_g          *
 *                              log(beta_G)             loglike_g       *
 * -------------------------------------------------------------------- *
 * Logistic                     alpha                   sens_alpha      *
 *                              b                       loc_l           *
 *                              log(alpha)              sens_l          *
 *                              log(b)                  logloc_l        *
 *                              b'                      relloc_l        *
 *                              beta_L                  like_l          *
 *                              log(beta_L)             loglike_l       *
 *                              B'_H                    hodos           *
 *                              B"                      grier           *
 * -------------------------------------------------------------------- *
 * Rectangular                  q                       sens_q          *
 *                              P(c)                    sens_pc         *
 *                              k                       loc_k           *
 *                              k'                      relloc_k        *
 * -------------------------------------------------------------------- *
 * "Nonparametric"              A'                      sens_np         *
 * -------------------------------------------------------------------- *
 * Other ones                   LOR                     sens_lor        *
 *                              Yule's Q                sens_yule       *
 * -------------------------------------------------------------------- *
 * ROC                          slope                   roc_slope       *
 *                              Delta_m                 delta_m         *
 *                              d_a                     roc_da          *
 *                              d'_e                    roc_de          *
 *                              A_z                     roc_az          *
 * -------------------------------------------------------------------- *
 * Print revision info                                  detect_ver      *
 * -------------------------------------------------------------------- *
 *                                                                      *
 * Copyright (C) Koen Van der Goten and Andre Vandierendonck.           *
 * Everyone is permitted to copy and distribute verbatim copies of this *
 * document, but changing is not allowed. The software is provided      *
 * as-is without any warranty of any kind. Selling or sublicensing of   *
 * this product is not allowed.                                         *
 * ==================================================================== *
 *}  

interface

const		{ a big constant }
   HUGE = 179769313486231570814527423731704356798070567525844996598917476803157260780028538760589558632766878171540458953514.0;

type
   selection_type = ( S_ALL, S_GAUSSIAN, S_LOGISTIC,
                      S_RECTANGULAR, S_OTHER, S_ROC, S_ILLEGAL );

procedure detect_ver;
function ratio( x, y : extended ; extreme : boolean ) : extended;
function odeh( p : extended ) : extended;

function sens_g( ns, n, h1, f1 : extended ; extreme : boolean ) : extended;
function loc_g( ns, n, h1, f1 : extended ; extreme : boolean ) : extended;
function relloc_g( ns, n, h1, f1 : extended ; extreme : boolean ) : extended;
function like_g( ns, n, h1, f1 : extended ; extreme : boolean ) : extended;
function loglike_g( ns, n, h1, f1 : extended ; extreme : boolean ) : extended;

function sens_l( ns, n, h1, f1 : extended ; extreme : boolean ) : extended;
function sens_alfa( ns, n, h1, f1 : extended ; extreme : boolean ) : extended;
function loc_l( ns, n, h1, f1 : extended ; extreme : boolean ) : extended;
function logloc_l( ns, n, h1, f1 : extended ; extreme : boolean ) : extended;
function relloc_l( ns, n, h1, f1 : extended ; extreme : boolean ) : extended;
function like_l( ns, n, h1, f1 : extended ; extreme : boolean ) : extended;
function loglike_l( ns, n, h1, f1 : extended ; extreme : boolean ) : extended;
function hodos( ns, n, h1, f1 : extended ; extreme : boolean ) : extended;
function grier( ns, n, h1, f1 : extended ;
		       extreme : boolean ) : extended;

function sens_q( ns, n, h1, f1 : extended ; extreme : boolean ) : extended;
function sens_pc( ns, n, h1, f1 : extended ; extreme : boolean ) : extended;
function loc_k( ns, n, h1, f1 : extended ; extreme : boolean ) : extended;
function relloc_k( ns, n, h1, f1 : extended ; extreme : boolean ) : extended;
function sens_lor( ns, n, h1, f1 : extended ; extreme : boolean ) : extended;
function sens_yule( ns, n, h1, f1 : extended ; extreme : boolean ) : extended;
function sens_np( ns, n, h1, f1 : extended ; extreme : boolean ) : extended;
function roc_slope( ns, n, h1, f1, ns2, n2, h2, f2 : extended ;
		    extreme : boolean ) : extended;
function delta_m( ns, n, h1, f1, ns2, n2, h2, f2 : extended ;
		  extreme : boolean ) : extended;
function roc_da( ns, n, h1, f1, ns2, n2, h2, f2 : extended ;
	       extreme : boolean ) : extended;
function roc_de( ns, n, h1, f1, ns2, n2, h2, f2 : extended ;
	       extreme : boolean ) : extended;
function roc_az( ns, n, h1, f1, ns2, n2, h2, f2 : extended ;
	       extreme : boolean ) : extended;

implementation

procedure detect_ver;
begin
    writeln('Module detect: $Revision: 1.0 $, $Date: 1996/07/24 16:28:47 $')
end;

function ratio( x, y : extended ; extreme : boolean ) : extended;

   var
      r : extended;
   
   begin
      if ( x < 0.0 ) or ( y <= 0.0 ) or ( x > y ) then	{ invalid param }
      begin
	 writeln('Invalid parameter, exiting');
	 exit
      end;
      if extreme then					{ log-linear rule }
      begin
	 x := x + 0.5;		{ each entry of 2x2 contingency table +.5 }
	 y := y + 1.0		{ row and column totals +1. }
      end;

      r := x / y;

      if ( r = 0.0 ) then				{ 1/2n-rule }
	 r := 1.0 / ( 2.0 * y )
      else if ( r = 1.0 ) then
	 r := 1.0 - 1.0 / ( 2.0 * y );
      
      ratio := r;
   end;

	    { Approximation of the inverse normal distribution function }
	    { proposed by Odeh & Evans (1974) in Applied Statistics     }
	    { For discussion see Brophy (1985) in Behavior Research     }
	    { Methods, Instruments & Computers.                         }
   
function odeh( p : extended ) : extended;

   var
      r, y, z : extended;
   
   begin
      if ( p > 0.5 ) then
	 r := 1.0 - p
      else
	 r := p;
      y := sqrt( -2.0 * ln( r ));
      z := y -
	 ((((4.53642210148E-5 * y + 0.0204231210245) * y + 0.342242088547) * y
	       + 1) * y + 0.322232431088)
	 / ((((0.0038560700634 * y + 0.10353775285) * y + 0.531103462366) * y
	       + 0.588581570495) * y + 0.099348462606);
      if( p < 0.5 ) then
	 odeh := -z
      else
	 odeh := z
   end;

function sens_g( ns, n, h1, f1 : extended ; extreme : boolean ) : extended;
   
   var
      zh, zf, d : extended;

   begin
      h1 := ratio( h1, ns, extreme );
      f1 := ratio( f1, n, extreme );
      zh := odeh(h1);
      zf := odeh(f1);
      d := zh - zf;
      sens_g := d;
   end;

function loc_g( ns, n, h1, f1 : extended ; extreme : boolean ) : extended;

   var
      zh, zf, c : extended;

   begin
      h1 := ratio( h1, ns, extreme );
      f1 := ratio( f1, n, extreme );
      zh := odeh(h1);
      zf := odeh(f1);
      c := -0.5 * ( zf + zh );
      loc_g := c;
   end;

function relloc_g( ns, n, h1, f1 : extended ; extreme : boolean ) : extended;
   
   var
      d, c, cprime : extended;
   
   begin
      if ( h1 / ns ) = ( f1 / n ) then		{ measure is undefined }
	 relloc_g := HUGE
      else
      begin
	 d := sens_g( ns, n, h1, f1, extreme );
	 c := loc_g( ns, n, h1, f1, extreme );
	 cprime := c / d;
	 relloc_g := cprime
      end
   end;

function like_g( ns, n, h1, f1 : extended ; extreme : boolean ) : extended;

   var
      b, zh, zf, hh, hf : extended;
   
   begin
      if ( h1 / ns ) = ( f1 / n ) then		{ measure is undefined }
	 like_g := HUGE
      else
      begin
	 h1 := ratio( h1, ns, extreme );
	 f1 := ratio( f1, n, extreme );
	 zh := odeh( h1 );
	 zf := odeh( f1 );

	 hh := (1.0 / ( sqrt( 2.0 * pi )) * exp( -0.5 * zh * zh ) );
	 hf := (1.0 / ( sqrt( 2.0 * pi )) * exp( -0.5 * zf * zf ) );

	 b := ( hh / hf ) * ( ns / n );
	 if ( f1 > h1 ) then
	    like_g := 1.0 / b
	 else
	    like_g := b
      end
   end;

function loglike_g( ns, n, h1, f1 : extended ; extreme : boolean ) : extended;

   begin
      if ( h1 / ns ) = ( f1 / n ) then		{ measure is undefined }
	 loglike_g := HUGE
      else
	 loglike_g := ln( like_g( ns, n, h1, f1, extreme ) )
   end;

function sens_l( ns, n, h1, f1 : extended ; extreme : boolean ) : extended;

   begin
      h1 := ratio( h1, ns, extreme );
      f1 := ratio( f1, n, extreme );
      sens_l := 0.5 * ln( h1 / ( 1.0 - h1 )) - 0.5 * ln( f1 / ( 1.0 - f1 ));
   end;

function sens_alfa( ns, n, h1, f1 : extended ; extreme : boolean ) : extended;

   begin
      h1 := ratio( h1, ns, extreme );
      f1 := ratio( f1, n, extreme );
      sens_alfa := sqrt(( h1 / ( 1.0 - h1 )) / ( f1 / ( 1.0 - f1 )));
   end;

function loc_l( ns, n, h1, f1 : extended ; extreme : boolean ) : extended;

   begin
      h1 := ratio( h1, ns, extreme );
      f1 := ratio( f1, n, extreme );
      loc_l := sqrt(( 1.0 - h1 ) * ( 1.0 - f1 ) / ( h1 * f1 ));
   end;

function logloc_l( ns, n, h1, f1 : extended ; extreme : boolean ) : extended;

   begin
      h1 := ratio( h1, ns, extreme );
      f1 := ratio( f1, n, extreme );
      logloc_l := -0.5 * (ln( h1 / ( 1.0 - h1 )) + ln( f1 / ( 1.0 - f1 )));
   end;

function relloc_l( ns, n, h1, f1 : extended ; extreme : boolean ) : extended;

   var
      logalfa, logb, bprime : extended;

   begin
      if ( h1 / ns ) = ( f1 / n ) then		{ measure is undefined }
	 relloc_l := HUGE
      else
      begin
	 logalfa := sens_l( ns, n, h1, f1, extreme );
	 logb := logloc_l( ns, n, h1, f1, extreme );
	 bprime := logb / ( 2.0 * logalfa );
	 if ( logalfa < 0.0 ) then
	    relloc_l := -bprime
	 else
	    relloc_l := bprime
      end
   end;

function like_l( ns, n, h1, f1 : extended ; extreme : boolean ) : extended;

   begin
      if ( h1 / ns ) = ( f1 / n ) then		{ measure is undefined }
	 like_l := HUGE
      else
      begin
	 h1 := ratio( h1, ns, extreme );
	 f1 := ratio( f1, n, extreme );
	 if ( h1 > f1 ) then
	    like_l := h1 * ( 1.0 - h1 ) / ( f1 * ( 1.0 - f1 ))
	 else
	    like_l := f1 * ( 1.0 - f1 ) / ( h1 * ( 1.0 - h1 ))
      end
   end;

function loglike_l( ns, n, h1, f1 : extended ; extreme : boolean ) : extended;

   begin
      if ( h1 / ns ) = ( f1 / n ) then		{ measure is undefined }
	 loglike_l := HUGE
      else
	 loglike_l := ln( like_l( ns, n, h1, f1, extreme ) )
   end;

function hodos( ns, n, h1, f1 : extended ; extreme : boolean ) : extended;

   var
      x, y : extended;
   
   begin
      if ( h1 / ns ) = ( f1 / n ) then		{ measure is undefined }
	 hodos := HUGE
      else
      begin
	 h1 := ratio( h1, ns, extreme );
	 f1 := ratio( f1, n, extreme );
	 if ( h1 < f1 ) then
	 begin
	    x := h1;
	    y := f1
	 end
	 else
	 begin
	    x := f1;
	    y := h1
	 end;
	 if ( x + y <= 1.0 ) then
	    hodos := 1.0 - (x * (1.0 - x)) / (y * (1.0 - y))
	 else
	    hodos := (y * (1.0 - y)) / (x * (1.0 - x)) - 1.0
      end
   end;

function grier( ns, n, h1, f1 : extended ;
		       extreme : boolean ) : extended;

   var
      x, y : extended;

   begin
      if ( h1 / ns ) = ( f1 / n ) then		{ measure is undefined }
	 grier := HUGE
      else
      begin
	 h1 := ratio( h1, ns, extreme );
	 f1 := ratio( f1, n, extreme );
	 if ( h1 < f1 ) then
	 begin
	    x := h1;
	    y := f1
	 end
	 else
	 begin
	    x := f1;
	    y := h1
	 end;
	 grier := (y * (1.0 - y) - x * (1.0 - x))
			   / (y * (1.0 - y) + x * (1.0 - x));
      end
   end;

function sens_q( ns, n, h1, f1 : extended ; extreme : boolean ) : extended;

   begin
      h1 := ratio( h1, ns, extreme );
      f1 := ratio( f1, n, extreme );
      sens_q := (h1 - f1) / (1.0 - f1);
   end;

function sens_pc( ns, n, h1, f1 : extended ; extreme : boolean ) : extended;

   begin
      h1 := ratio( h1, ns, extreme );
      f1 := ratio( f1, n, extreme );
      sens_pc := 0.5 * (1.0 - (f1 - h1));
   end;

function loc_k( ns, n, h1, f1 : extended ; extreme : boolean ) : extended;

   begin
      h1 := ratio( h1, ns, extreme );
      f1 := ratio( f1, n, extreme );
      loc_k := 0.5 * (1.0 - (f1 + h1));
   end;

function relloc_k( ns, n, h1, f1 : extended ; extreme : boolean ) : extended;

   begin
      h1 := ratio( h1, ns, extreme );
      f1 := ratio( f1, n, extreme );
      relloc_k := 1.0 / (1.0 + (f1 / (1.0 - h1)));
   end;

function sens_lor( ns, n, h1, f1 : extended ; extreme : boolean ) : extended;

   begin
      h1 := ratio( h1, ns, extreme );
      f1 := ratio( f1, n, extreme );
      sens_lor := ln( (h1 * (1.0 - f1)) / (f1 * (1.0 - h1)));
   end;

function sens_yule( ns, n, h1, f1 : extended ; extreme : boolean ) : extended;

   begin
      h1 := ratio( h1, ns, extreme );
      f1 := ratio( f1, n, extreme );
      sens_yule := (h1 - f1) / (h1 - 2.0 * f1 * h1 + f1);
   end;

function sens_np( ns, n, h1, f1 : extended ; extreme : boolean ) : extended;

   var
      x, y : extended;

   begin
      y := ratio( h1, ns, extreme );
      x := ratio( f1, n, extreme );
      if( x <= y ) then
	 sens_np := 0.5 + (y - x + (x - y) * (x - y)) / (4.0 * y * (1.0 - x))
      else
	 sens_np := 0.5 - (x - y + (x - y) * (x - y)) / (4.0 * x * (1.0 - y))
   end;

function roc_slope( ns, n, h1, f1, ns2, n2, h2, f2 : extended ;
		    extreme : boolean ) : extended;

   var
      zh1, zf1, zh2, zf2 : extended;
   
   begin
      h1 := ratio( h1, ns, extreme );
      f1 := ratio( f1, n, extreme );
      h2 := ratio( h2, ns2, extreme );
      f2 := ratio( f2, n2, extreme );
      zh1 := odeh( h1 );
      zf1 := odeh( f1 );
      zh2 := odeh( h2 );
      zf2 := odeh( f2 );
      roc_slope := abs((zh2 - zh1) / (zf2 - zf1));
   end;

function delta_m( ns, n, h1, f1, ns2, n2, h2, f2 : extended ;
		  extreme : boolean ) : extended;

   var
      zh1, zf1, zh2, zf2, s : extended;
   
   begin
      h1 := ratio( h1, ns, extreme );
      f1 := ratio( f1, n, extreme );
      h2 := ratio( h2, ns2, extreme );
      f2 := ratio( f2, n2, extreme );
      zh1 := odeh( h1 );
      zf1 := odeh( f1 );
      zh2 := odeh( h2 );
      zf2 := odeh( f2 );
      s := abs((zh2 - zh1) / (zf2 - zf1));
      delta_m := zh1 / s - zf1;
   end;

function roc_da( ns, n, h1, f1, ns2, n2, h2, f2 : extended ;
	       extreme : boolean ) : extended;

   var
      zh1, zf1, zh2, zf2, s : extended;

   begin
      h1 := ratio( h1, ns, extreme );
      f1 := ratio( f1, n, extreme );
      h2 := ratio( h2, ns2, extreme );
      f2 := ratio( f2, n2, extreme );
      zh1 := odeh( h1 );
      zf1 := odeh( f1 );
      zh2 := odeh( h2 );
      zf2 := odeh( f2 );
      s := abs((zh2 - zh1) / (zf2 - zf1));
      roc_da := sqrt( 2.0 / (1.0 + s * s)) * (zh1 - s * zf1);
   end;

function roc_de( ns, n, h1, f1, ns2, n2, h2, f2 : extended ;
	       extreme : boolean ) : extended;

   var
      zh1, zf1, zh2, zf2, s : extended;
   
   begin
      h1 := ratio( h1, ns, extreme );
      f1 := ratio( f1, n, extreme );
      h2 := ratio( h2, ns2, extreme );
      f2 := ratio( f2, n2, extreme );
      zh1 := odeh( h1 );
      zf1 := odeh( f1 );
      zh2 := odeh( h2 );
      zf2 := odeh( f2 );
      s := abs((zh2 - zh1) / (zf2 - zf1));
      roc_de := 2.0 / (1.0 + s) * (zh1 - s * zf1);
   end;

function roc_az( ns, n, h1, f1, ns2, n2, h2, f2 : extended ;
	       extreme : boolean ) : extended;

   var
      da : extended;
   
   begin
      da := roc_da( ns, n, h1, f1, ns2, n2, h2, f2, extreme );
      roc_az := (1.0 / sqrt( 2.0 * pi )) * exp( -0.5 * da * da );
   end;
end.
